﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using Trinity.Cache;
using Trinity.Combat.Abilities;
using Trinity.Configuration;
using Trinity.DbProvider;
using Trinity.Helpers;
using Trinity.Items;
using Trinity.Reference;
using Trinity.Technicals;
using Trinity.UI.UIComponents;
using Zeta.Bot;
using Zeta.Common;
using Zeta.Common.Plugins;
using Zeta.Game;
using Zeta.Game.Internals;
using Zeta.Game.Internals.Actors;
using Zeta.Game.Internals.SNO;
using Logger = Trinity.Technicals.Logger;

namespace Trinity
{
    public partial class Trinity : IPlugin
    {
        private static int Tick = 0;
        /// <summary>
        /// For backwards compatability
        /// </summary>
        public static void RefreshDiaObjects()
        {
            // Framelock should happen in the MainLoop, where we read all the actual ACD's
            using (new MemorySpy("RefreshDiaObjects()"))
            {
                RefreshDiaObjectCache();
            }
        }

        /// <summary>
        /// This method will add and update necessary information about all available actors. Determines ObjectType, 
        /// sets ranges, updates blacklists, determines avoidance, kiting, target weighting
        /// and the result is we will have a new target for the Target Handler. Returns true if the cache was refreshed.
        /// </summary>
        /// <returns>True if the cache was updated</returns>
        public static bool RefreshDiaObjectCache()
        {
            if (Tick > 10) Tick = 0;
            Tick++;

            if (!Player.IsFullyValid)
                return false;

            MonkCombat.RunOngoingPowers();

            using (new PerformanceLogger("RefreshDiaObjectCache"))
            {
                /* Reduce ignore-for-loops counter */
                if (IgnoredRActorGuidForLoops > 0)
                    IgnoredRActorGuidForLoops--;

                /* Refresh the Cache */
                using (new PerformanceLogger("RefreshDiaObjectCache.UpdateBlock"))
                {
                    /* Refresh at Tick pair */
                    if ((Tick & 1) != 1)
                    {
                        using (new MemorySpy("RefreshDiaObjects().Maintain"))
                        {
                            GenericCache.MaintainCache();
                            GenericBlacklist.MaintainBlacklist();
                            CacheData.Clear();
                        }

                        /* Now pull up all the data and store anything we want to handle in the super special cache list
                        Also use many cache dictionaries to minimize DB<->D3 memory hits, and speed everything up a lot */
                        using (new MemorySpy("RefreshDiaObjects().Loop"))
                        {
                            RefreshCacheInit();
                            RefreshCacheMainLoop();
                            RefreshCacheMarkers();
                        }

                        /* Add Team HotSpots to the cache */
                        using (new MemorySpy("RefreshDiaObjects().HotSpots"))
                        {
                            foreach (var ghs in GroupHotSpots.GetCacheObjectHotSpots())
                            {
                                ObjectCache.Add(ghs);
                            }
                        }

                        using (new MemorySpy("RefreshDiaObjects().Avoidance"))
                        { 
                            RefreshKiteValue();
                            #region FireChains
                            if (Settings.Combat.Misc.UseExperimentalFireChainsAvoidance)
                            {
                                const float fireChainSize = 5f;
                                foreach (var unit1 in ObjectCache.Where(u => u.MonsterAffixes.HasFlag(MonsterAffixes.FireChains)))
                                {
                                    foreach (var unit2 in ObjectCache.Where(u => u.MonsterAffixes.HasFlag(MonsterAffixes.FireChains)).Where(unit2 => unit1.RActorGuid != unit2.RActorGuid))
                                    {
                                        for (float i = 0; i <= unit1.Position.Distance2D(unit2.Position); i += (fireChainSize / 4))
                                        {
                                            Vector3 fireChainSpot = MathEx.CalculatePointFrom(unit1.Position, unit2.Position, i);

                                            if (Player.Position.Distance2D(fireChainSpot) <= fireChainSize)
                                            {
                                                Logger.Log(TrinityLogLevel.Debug, LogCategory.CacheManagement, "Avoiding Fire Chains!");
                                                Trinity.Player.StandingInAvoidance = true;
                                            }
                                            CacheData.AvoidanceObstacles.Add(new CacheObstacleObject(fireChainSpot, fireChainSize, -2, "FireChains"));
                                        }
                                    }
                                    if (CacheData.AvoidanceObstacles.Any(aoe => aoe.ActorSNO == -2))
                                        Logger.Log(TrinityLogLevel.Debug, LogCategory.CacheManagement, "Generated {0} avoidance points for FireChains, minDistance={1} maxDistance={2}",
                                            CacheData.AvoidanceObstacles.Count(aoe => aoe.ActorSNO == -2),
                                            CacheData.AvoidanceObstacles.Where(aoe => aoe.ActorSNO == -2)
                                                .Min(aoe => aoe.Position.Distance2D(Trinity.Player.Position)),
                                            CacheData.AvoidanceObstacles.Where(aoe => aoe.ActorSNO == -2)
                                                .Max(aoe => aoe.Position.Distance2D(Trinity.Player.Position)));
                                }
                            }
                            #endregion
                        }

                        using (new MemorySpy("RefreshDiaObjects().CacheData"))
                        {
                            CacheData.ClearObsolete();
                            CacheData.SetDictionary();
                        }

                        using (new MemorySpy("RefreshDiaObjects().CombatContext"))
                        {
                            CombatBase.RefreshValues();
                            RefreshDiaGetWeights();
                        }

                        /* Invoke all methode called in cache update */
                        using (new MemorySpy("RefreshDiaObjects().InvokeEvents"))
                        {
                            Events.OnCacheUpdatedHandler.Invoke();
                        }

                        /*using (new MemorySpy("RefreshDiaObjects().CacheUI"))
                        {
                            CacheUI.DataModel.SourceCacheObjects.Clear(); 
                            foreach (var co in ObjectCache)
                            {
                                CacheUI.DataModel.SourceCacheObjects.Add(co.Copy());
                            }
                        }*/

                    }
                    /* Refresh at Tick impair */
                    else
                    {
                        using (new MemorySpy("RefreshDiaObjects().Kite"))
                        { 
                            RefreshKiteValue(); 
                        }

                        /* Refresh grid map fields */
                        using (new MemorySpy("RefreshDiaObjects().Grid"))
                        {
                            GridMap.RefreshGridMainLoop();
                        }
                    }
                }

                using (new MemorySpy("RefreshDiaObjects().FinalCheck"))
                {
                    if (ObjectCache != null)
                    {
                        // Pre-townrun is too far away
                        if (!Player.IsInTown && TownRun.PreTownRunPosition != Vector3.Zero && TownRun.PreTownRunWorldId == Player.WorldID && !ForceVendorRunASAP
                            && TownRun.PreTownRunPosition.Distance2D(Player.Position) <= V.F("Cache.PretownRun.MaxDistance"))
                        {
                            Logger.Log(TrinityLogLevel.Debug, LogCategory.UserInformation,
                                "Pre-TownRun position is more than {0} yards away, canceling", V.I("Cache.PretownRun.MaxDistance"));
                            TownRun.PreTownRunPosition = Vector3.Zero;
                            TownRun.PreTownRunWorldId = -1;
                        }

                        // Reached pre-townrun position
                        if (!Player.IsInTown && TownRun.PreTownRunPosition != Vector3.Zero && TownRun.PreTownRunWorldId == Player.WorldID && !ForceVendorRunASAP
                            && TownRun.PreTownRunPosition.Distance2D(Player.Position) <= 15f)
                        {
                            Logger.Log(TrinityLogLevel.Debug, LogCategory.UserInformation, "Successfully returned to Pre-TownRun Position");
                            TownRun.PreTownRunPosition = Vector3.Zero;
                            TownRun.PreTownRunWorldId = -1;
                        }

                        // After a townrun, make sure to return to original TownRun Location
                        if (!Player.IsInTown && CurrentTarget == null && TownRun.PreTownRunPosition != Vector3.Zero && TownRun.PreTownRunWorldId == Player.WorldID && !ForceVendorRunASAP)
                        {
                            if (TownRun.PreTownRunPosition.Distance2D(Player.Position) > 10f && TownRun.PreTownRunPosition.Distance2D(Player.Position) <= V.F("Cache.PretownRun.MaxDistance"))
                            {
                                CurrentTarget = new TrinityCacheObject()
                                {
                                    Position = TownRun.PreTownRunPosition,
                                    Type = GObjectType.Player,
                                    Weight = 20000,
                                    Distance = 2f,
                                    Radius = 2f,
                                    InternalName = "PreTownRunPosition"
                                };
                                Logger.Log(TrinityLogLevel.Debug, LogCategory.UserInformation, "Returning to Pre-TownRun Position");
                            }
                        }

                        using (new PerformanceLogger("RefreshDiaObjectCache.FinalChecks"))
                        {
                            if (Settings.WorldObject.EnableBountyEvents)
                            {
                                bool eventObjectNear = ObjectCache.Any(o => o.Type == GObjectType.CursedChest || o.Type == GObjectType.CursedShrine);

                                if (!Player.InActiveEvent)
                                {
                                    EventStartPosition = Vector3.Zero;
                                    EventStartTime = DateTime.MinValue;
                                }

                                // Reset Event time while we have targts
                                if (CurrentTarget != null && Player.InActiveEvent && eventObjectNear)
                                {
                                    EventStartTime = DateTime.UtcNow;
                                }

                                if (eventObjectNear)
                                {
                                    EventStartPosition = ObjectCache.FirstOrDefault(o => o.Type == GObjectType.CursedChest || o.Type == GObjectType.CursedShrine).Position;
                                }

                                var activeEvent = ZetaDia.ActInfo.ActiveQuests.FirstOrDefault(q => ((SNOQuest)q.QuestSNO).ToString().Contains("Event"));

                                const int waitTimeoutSeconds = 90;
                                if (DateTime.UtcNow.Subtract(EventStartTime).TotalSeconds > waitTimeoutSeconds && activeEvent != null)
                                {
                                    CacheData.BlacklistedEvents.Add(activeEvent.QuestSNO);
                                }

                                if (CurrentTarget == null && Player.InActiveEvent && EventStartPosition != Vector3.Zero &&
                                    DateTime.UtcNow.Subtract(EventStartTime).TotalSeconds < waitTimeoutSeconds &&
                                    activeEvent != null && !CacheData.BlacklistedEvents.Contains(activeEvent.QuestSNO))
                                {
                                    CurrentTarget = new TrinityCacheObject()
                                    {
                                        Position = EventStartPosition,
                                        Type = GObjectType.Player,
                                        Weight = 20000,
                                        Distance = 2f,
                                        Radius = 2f,
                                        InternalName = "WaitForEvent"
                                    };
                                    Logger.Log("Waiting for Event - Time Remaining: {0:0} seconds",
                                        waitTimeoutSeconds - DateTime.UtcNow.Subtract(EventStartTime).TotalSeconds);
                                }
                            }

                            if (CurrentTarget == null)
                            {
                                RefreshWaitTimers();
                                CombatBase.SwitchToTarget(CurrentTarget);
                            }

                            // Store less important objects.
                            if (ObjectCache.Count > 1)
                            {
                                var setting = Settings.Advanced.CacheWeightThresholdPct;
                                var threshold = setting > 0 && CurrentTarget != null ? CurrentTarget.Weight * ((double)setting / 100) : 0;

                                var lowPriorityObjects = ObjectCache.DistinctBy(c => c.RActorGuid).Where(c =>
                                    c.Type != GObjectType.Avoidance && c.Type != GObjectType.Unit ||
                                    c.Weight < threshold && c.Distance > 12f && !c.IsElite
                                    ).ToDictionary(x => x.RActorGuid, x => x);

                                Logger.Log(TrinityLogLevel.Debug, LogCategory.CacheManagement, "Cached {0}/{1} ({2:0}%) WeightThreshold={3}",
                                    lowPriorityObjects.Count,
                                    ObjectCache.Count,
                                    lowPriorityObjects.Count > 0 ? ((double)lowPriorityObjects.Count / ObjectCache.Count) * 100 : 0,
                                    threshold);

                                CacheData.LowPriorityObjectCache = lowPriorityObjects;
                            }
                        }
                    }
                }

                // We have a target and the cached was refreshed
                return CurrentTarget != null;
            }
        }

        /// <summary>
        /// Adds Legendary & Set Minimap Markers to ObjectCache
        /// </summary>
        private static void RefreshCacheMarkers()
        {
            const int setItemMarkerTexture = 404424;
            const int legendaryItemMarkerTexture = 275968;

            if (!WantToTownRun && !ForceVendorRunASAP)
            {
                var legendaryItemMarkers = ZetaDia.Minimap.Markers.CurrentWorldMarkers.Where(m => m.IsValid &&
                                    m.Position.Distance2D(Player.Position) >= 15f &&
                                    m.Position.Distance2D(Player.Position) < 495f &&
                                    (m.MinimapTexture == setItemMarkerTexture || m.MinimapTexture == legendaryItemMarkerTexture)).ToList();

                foreach (var marker in legendaryItemMarkers)
                {
                    var name = (marker.MinimapTexture == setItemMarkerTexture ? "Set Item" : "Legendary Item") + " Minimap Marker";
                    var hash = marker.NameHash + marker.Position.ToString();

                    if (GenericBlacklist.ContainsKey(hash))
                    {
                        Logger.Log(TrinityLogLevel.Info, LogCategory.UserInformation, "Ignoring Marker because it's blacklisted {0} {1} at {2} distance {3}",
                            name, marker.NameHash, marker.Position, marker.Position.Distance(Player.Position));
                        continue;
                    }

                    Logger.LogDebug(LogCategory.CacheManagement, "Adding {0} {1} at {2} distance {3}", name, marker.NameHash, marker.Position, marker.Position.Distance(Player.Position));
                    ObjectCache.Add(new TrinityCacheObject()
                    {
                        Position = marker.Position,
                        InternalName = name,
                        RActorGuid = marker.NameHash,
                        ActorType = ActorType.Item,
                        Type = GObjectType.Item,
                        ItemQuality = ItemQuality.Legendary,
                        ObjectHash = hash,
                        Radius = 2f,
                        Weight = 50,
                        IsMarker = true
                    });
                }

                if (legendaryItemMarkers.Any() && TrinityItemManager.FindValidBackpackLocation(true) != new Vector2(-1, -1))
                {
                    var legendaryItems = ZetaDia.Actors.GetActorsOfType<DiaItem>().Where(i => i.IsValid && i.IsACDBased && i.Position.Distance2D(ZetaDia.Me.Position) < 5f &&
                        legendaryItemMarkers.Any(im => i.Position.Distance2D(i.Position) < 2f));

                    foreach (var diaItem in legendaryItems)
                    {
                        Logger.LogDebug(LogCategory.CacheManagement, "Adding Legendary Item from Marker {0} dist={1} ActorSNO={2} ACD={3} RActor={4}",
                            diaItem.Name, diaItem.Distance, diaItem.ActorSNO, diaItem.ACDGuid, diaItem.RActorGuid);

                        ObjectCache.Add(new TrinityCacheObject()
                        {
                            Position = diaItem.Position,
                            InternalName = diaItem.Name,
                            RActorGuid = diaItem.RActorGuid,
                            ActorSNO = diaItem.ActorSNO,
                            ACDGuid = diaItem.ACDGuid,
                            IsInLineOfSight = true,
                            Distance = diaItem.Distance,
                            ActorType = ActorType.Item,
                            Type = GObjectType.Item,
                            Radius = 2f,
                            Weight = 50,
                            ItemQuality = ItemQuality.Legendary,
                            RequiredRange = 1f,
                        });
                    }
                }
            }

            bool isRiftGuardianQuestStep = ZetaDia.CurrentQuest.QuestSNO == 337492 && ZetaDia.CurrentQuest.StepId == 16;

            if (isRiftGuardianQuestStep)
            {
                // Add Rift Guardian POI's or Markers to ObjectCache
                const int riftGuardianMarkerTexture = 81058;
                Func<MinimapMarker, bool> riftGuardianMarkerFunc = m => m.IsValid && (m.IsPointOfInterest || m.MinimapTexture == riftGuardianMarkerTexture) && m.Position.Distance2D(Player.Position) < 495f;

                foreach (var marker in ZetaDia.Minimap.Markers.CurrentWorldMarkers.Where(riftGuardianMarkerFunc))
                {
                    Logger.LogDebug(LogCategory.CacheManagement, "Adding Rift Guardian POI, distance {0}", marker.Position.Distance2D(Player.Position));
                    ObjectCache.Add(new TrinityCacheObject()
                    {
                        Position = marker.Position,
                        InternalName = "Rift Guardian",
                        RActorGuid = marker.NameHash,
                        ActorType = ActorType.Monster,
                        Type = GObjectType.Unit,
                        Radius = 35f,
                        Weight = 5000,
                        IsBoss = true,
                    });
                }
            }

            if (isRiftGuardianQuestStep || Player.ParticipatingInTieredLootRun) // X1_LR_DungeonFinder
            {
                foreach (var marker in ZetaDia.Minimap.Markers.CurrentWorldMarkers.Where(m => m.IsPointOfInterest))
                {
                    ObjectCache.Add(new TrinityCacheObject()
                    {
                        Position = marker.Position,
                        InternalName = "Rift Guardian",
                        RActorGuid = marker.NameHash,
                        ActorType = ActorType.Monster,
                        Type = GObjectType.Unit,
                        Radius = 10f,
                        Weight = 5000,
                        IsBoss = true,
                    });
                }
            }
        }

        private static void RefreshCacheMainLoop()
        {
            using (new PerformanceLogger("CacheManagement.RefreshCacheMainLoop"))
            {
                IEnumerable<DiaObject> refreshSource;

                if (Settings.Advanced.LogCategories.HasFlag(LogCategory.CacheManagement))
                {
                    refreshSource = ReadDebugActorsFromMemory();
                }
                else
                {
                    refreshSource = ReadActorsFromMemory();
                }
                Stopwatch t1 = new Stopwatch();

                foreach (DiaObject currentObject in refreshSource)
                {
                    try
                    {
                        // Objects deemed of low importance are stored from the last refresh
                        TrinityCacheObject cachedObject;
                        if (CacheData.LowPriorityObjectCache.TryGetValue(currentObject.RActorGuid, out cachedObject))
                        {
                            cachedObject.Distance = currentObject.Distance;
                            var timeSinceRefresh = DateTime.UtcNow.Subtract(cachedObject.LastSeenTime).TotalMilliseconds;

                            // Determine if we should use the stored object or not
                            if (timeSinceRefresh < Settings.Advanced.CacheLowPriorityRefreshRate && cachedObject.Distance > 12f)
                            {
                                cachedObject.Position = currentObject.Position;
                                ObjectCache.Add(cachedObject);
                                continue;
                            }
                        }

                        if (!Settings.Advanced.LogCategories.HasFlag(LogCategory.CacheManagement))
                        {
                            /*
                             *  Main Cache Function
                             */
                            CacheDiaObject(currentObject);
                        }
                        else
                        {
                            // We're debugging, slightly slower, calculate performance metrics and dump debugging to log 
                            t1.Reset();
                            t1.Start();

                            /*
                             *  Main Cache Function
                             */
                            bool addToCache = CacheDiaObject(currentObject);

                            if (t1.IsRunning)
                                t1.Stop();

                            double duration = t1.Elapsed.TotalMilliseconds;

                            // don't log stuff we never care about
                            if (duration <= 1 && (c_IgnoreReason == "IgnoreName" || c_IgnoreReason == "InvalidName"))
                                continue;

                            string extraData = "";

                            switch (CurrentCacheObject.Type)
                            {
                                case GObjectType.Unit:
                                    {
                                        if (c_IsEliteRareUnique)
                                            extraData += " IsElite " + c_MonsterAffixes.ToString();

                                        if (c_unit_HasShieldAffix)
                                            extraData += " HasAffixShielded";

                                        if (c_HasDotDPS)
                                            extraData += " HasDotDPS";

                                        if (CurrentCacheObject.IsInLineOfSight)
                                            extraData += " HasBeenInLoS";

                                        extraData += " HP=" + c_HitPoints.ToString("0") + " (" + c_HitPointsPct.ToString("0.00") + ")";
                                    } break;
                                case GObjectType.Avoidance:
                                    {
                                        extraData += Trinity.Player.StandingInAvoidance ? "InAoE " : "";
                                        break;
                                    }
                            }

                            var ll = TrinityLogLevel.Debug;
                            //if (CurrentCacheObject.IsUnit && !addToCache && !CurrentCacheObject.IsSummonedByPlayer && CurrentCacheObject.HitPoints > 0)
                                //ll = TrinityLogLevel.Error;

                            if (c_IgnoreReason != "InternalName")
                                Logger.Log(ll, LogCategory.CacheManagement,
                                    "[{0:0000.00}ms] {1} {2} Type: {3} ({5}) Name={6} ({7}) {8} {9} Dist2Mid={10:0} Dist2Rad={11:0} ZDiff={12:0} Radius={13:0} RAGuid={14} {15}",
                                    duration,
                                    (addToCache ? "Added " : "Ignored"),
                                    (!addToCache ? ("By: " + (c_IgnoreReason != "None" ? c_IgnoreReason : "") + (c_InfosSubStep != "" ? c_InfosSubStep : "")) : ""),
                                    CurrentCacheObject.ActorType,
                                    CurrentCacheObject.GizmoType != GizmoType.None ? CurrentCacheObject.GizmoType.ToString() : "",
                                    CurrentCacheObject.Type,
                                    CurrentCacheObject.InternalName,
                                    CurrentCacheObject.ActorSNO,
                                    (CurrentCacheObject.IsBoss ? " IsBoss" : ""),
                                    (c_CurrentAnimation != SNOAnim.Invalid ? " Anim: " + c_CurrentAnimation : ""),
                                    CurrentCacheObject.Distance,
                                    CurrentCacheObject.RadiusDistance,
                                    c_ZDiff,
                                    CurrentCacheObject.Radius,
                                    CurrentCacheObject.RActorGuid,
                                    extraData);
                        }
                    }
                    catch (Exception ex)
                    {
                        TrinityLogLevel ll = TrinityLogLevel.Debug;
                        LogCategory lc = LogCategory.CacheManagement;

                        if (ex is NullReferenceException)
                        {
                            ll = TrinityLogLevel.Error;
                            lc = LogCategory.UserInformation;
                        }

                        string gizmoType = "";
                        var giz = currentObject as DiaGizmo;
                        if (giz != null && giz.CommonData.IsValid)
                        {
                            gizmoType = "GizmoType: " + giz.CommonData.ActorInfo.GizmoType.ToString();
                        }
                        Logger.Log(ll, lc, "Error while refreshing DiaObject ActorSNO: {0} Name: {1} Type: {2} Distance: {3:0} {4} {5}",
                                currentObject.ActorSNO, currentObject.Name, currentObject.ActorType, currentObject.Distance, gizmoType, ex.Message);

                    }
                }

            }
        }

        // Refresh object list from Diablo 3 memory RefreshDiaObjects()
        private static void RefreshCacheInit()
        {
            using (new PerformanceLogger("RefreshDiaObjectCache.CacheInit"))
            {
                if (LastTarget == null)
                    LastTarget = new TrinityCacheObject() { RActorGuid = -1, };

                LastTarget = CurrentTarget ?? LastTarget;
                CurrentTarget = null;

                KillRange = Settings.Combat.Misc.NonEliteRange;
                if (DataDictionary.QuestLevelAreaIds.Contains(Player.LevelAreaId) && CombatBase.IsQuestingMode)
                    KillRange = 300f;

                if (!ProfileManager.CurrentProfile.KillMonsters || !CombatTargeting.Instance.AllowedToKillMonsters)
                    KillRange = 0;

                LootRange = Zeta.Bot.Settings.CharacterSettings.Instance.LootRadius;
                if (!ProfileManager.CurrentProfile.PickupLoot || !LootTargeting.Instance.AllowedToLoot)
                    LootRange = 0;

                AnyMobsInRange = false;
                AnyTreasureGoblinsPresent = false;

                PlayerOwnedMysticAllyCount = 0;
                PlayerOwnedGargantuanCount = 0;
                PlayerOwnedZombieDogCount = 0;
                PlayerOwnedFetishCount = 0;
                PlayerOwnedDHPetsCount = 0;
                PlayerOwnedDHSentryCount = 0;
                PlayerOwnedHydraCount = 0;

                ShouldStayPutDuringAvoidance = false;

                Player.StandingInAvoidance = false;
                Player.TryToAvoidProjectile = false;
                Player.NeedToKite = false;

                ObjectCache.Clear();
            }
        }

        private static void ClearCachesOnGameChange(object sender, EventArgs e)
        {
            CacheData.FullClear();
        }

        private static List<DiaObject> ReadDebugActorsFromMemory()
        {
            return (from o in ZetaDia.Actors.GetActorsOfType<DiaObject>(true, false)
                    where o.IsValid && o.CommonData != null && o.CommonData.IsValid
                    orderby o.Distance
                    select o).ToList();
        }

        private static IEnumerable<DiaObject> ReadActorsFromMemory()
        {
            return from o in ZetaDia.Actors.GetActorsOfType<DiaObject>(true, false)
                   where o.IsValid && o.CommonData != null && o.CommonData.IsValid
                   select o;
        }

        private static void RefreshWaitTimers()
        {

            // See if we should wait for [playersetting] milliseconds for possible loot drops before continuing run
            if (CurrentTarget == null &&
                (DateTime.UtcNow.Subtract(LastHadUnitInSights).TotalMilliseconds <= Settings.Combat.Misc.DelayAfterKill ||
                DateTime.UtcNow.Subtract(LastHadEliteUnitInSights).TotalMilliseconds <= Settings.Combat.Misc.DelayAfterKill ||
                DateTime.UtcNow.Subtract(LastHadBossUnitInSights).TotalMilliseconds <= 3000 ||
                DateTime.UtcNow.Subtract(Helpers.Composites.LastFoundHoradricCache).TotalMilliseconds <= 5000) ||
                DateTime.UtcNow.Subtract(LastHadContainerInSights).TotalMilliseconds <= Settings.WorldObject.OpenContainerDelay)
            {
                CurrentTarget = new TrinityCacheObject()
                {
                    Position = Player.Position,
                    Type = GObjectType.Player,
                    Weight = 20000,
                    Distance = 2f,
                    Radius = 2f,
                    InternalName = "WaitForLootDrops"
                };
                Logger.Log(TrinityLogLevel.Debug, LogCategory.Behavior, "Waiting for loot to drop, delay: {0}ms", Settings.Combat.Misc.DelayAfterKill);
            }
        }
    }
}
