﻿using System;
using System.Globalization;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Trinity.Combat;
using Trinity.Combat.Abilities;
using Trinity.Config.Combat;
using Trinity.DbProvider;
using Trinity.Helpers;
using Trinity.Items;
using Trinity.Technicals;
using Zeta.Bot;
using Zeta.Bot.Navigation;
using Zeta.Common;
using Zeta.Game;
using Zeta.Game.Internals.Actors;
using Zeta.Game.Internals.SNO;
using Zeta.TreeSharp;
using Logger = Trinity.Technicals.Logger;

namespace Trinity
{
    public partial class Trinity
    {
        /// <summary>
        /// Handles all aspects of moving to and attacking the current target
        /// </summary>
        /// <returns></returns>
        internal static async Task<RunStatus> HandleTarget()
        {
            using (new MemorySpy("HandleTarget()"))
            {
                try
                {
                    if (!Player.IsInGame)
                        return GetRunStatus("NotInGame", RunStatus.Failure);

                    if (Player.IsLoadingWorld)
                        return GetRunStatus("LoadingWorld", RunStatus.Failure);

                    if (!Player.IsValid)
                        return GetRunStatus("MeInvalid", RunStatus.Failure);

                    if (!Player.CommonData.IsValid)
                        return GetRunStatus("CDInvalid", RunStatus.Failure);

                    if (Player.IsDead)
                        return GetRunStatus("HeroIsDead", RunStatus.Failure);

                    if (UsePotionIfNeededTask()) 
                        return GetRunStatus("UsePotion", RunStatus.Running);

                    CheckAvoidDeath();
                    HandleAssignAbilityTask();
                    SetTargetFields();

                    if (HandleTownRunWaitTask())      
                        return GetRunStatus("WaitForTownRun", RunStatus.Running);

                    if (HandleTownRunReadyTask())          
                        return GetRunStatus("TownRunReady", RunStatus.Success);

                    if (CurrentTarget == null)
                        return GetRunStatus("TargetNull", RunStatus.Failure);

                    if (HandleCastingOrLoadingTask())
                        return GetRunStatus("TownPortal", RunStatus.Success);

                    PositionCache.AddPosition();

                    if (HandleObjectInRange())
                        return GetRunStatus("HandleObject", RunStatus.Running);

                    UpdateStatusTextTarget(false);

                    if (Player.IsIncapacitated || Player.IsRooted)
                        return GetRunStatus("HeroRooted", RunStatus.Running);

                    if (UsedSpecialMovement())
                        return GetRunStatus("SpecialMovement", RunStatus.Running);

                    SetQueuedSpecialMovement();
                    SetQueuedBasicMovement();

                    if (CombatBase.QueuedMovement.IsQueuedMovement)
                        await CombatBase.QueuedMovement.Execute();

                    return GetRunStatus("EndLoop", RunStatus.Running);
                }
                catch (Exception ex)
                {
                    Logger.LogError("Error in HandleTarget: {0}", ex);
                    return GetRunStatus("Error", RunStatus.Running);
                }
            }
        }

        /// <summary>
        /// Returns a RunStatus
        /// </summary>
        /// <param name="rs"></param>
        /// <returns></returns>
        private static RunStatus GetRunStatus(string post, RunStatus status)
        {
            Logger.Log(TrinityLogLevel.Debug, LogCategory.Targetting, "| HandleTarget returning {0} reason:{1}", status, post);

            if (Trinity.Settings.Advanced.LogCategories.HasFlag(LogCategory.Targetting))
            {
                if (CurrentTarget != null)
                {
                    Logger.Log(TrinityLogLevel.Debug, LogCategory.Targetting,
                        "| =>   CurrentTarget  : {0}",
                        CurrentTarget.Infos);
                }

                if (!CombatBase.IsNull(CombatBase.CurrentPower))
                {
                    Logger.Log(TrinityLogLevel.Debug, LogCategory.Targetting,
                        "| =>   CurrentPower   : {0}", CombatBase.CurrentPower.SNOPower + " InRange:" + TargetIsInRange + " Dist:" + TargetDistance);
                    Logger.Log(TrinityLogLevel.Debug, LogCategory.Targetting,
                        "| =>   MCD:{0} TCD{1} TCDi:{2} CPTP:{3} CPMP:{4} ACD:{5}",
                        MovePosition, TargetPosition, TargetDistance, CombatBase.CurrentPower.TargetPosition, CombatBase.CurrentPower.MovePosition, CombatBase.CurrentPower.TargetACDGUID);
                }

                if (CombatBase.QueuedMovement.IsQueuedMovement)
                {
                    Logger.Log(TrinityLogLevel.Debug, LogCategory.Targetting,
                        "| =>   QueuedMovement : {0}",
                        "Type: " + CombatBase.QueuedMovement.CurrentMovement.Options.Type + " Infos: " + CombatBase.QueuedMovement.CurrentMovement.Infos);
                }

                if (ObjectCache != null)
                {
                    Logger.Log(TrinityLogLevel.Debug, LogCategory.Targetting,
                        "| =>   ObjectCache    : {0}",
                        ObjectCache.Count().ToString() + " objects");
                }

                if (MainGrid.Map.Any() && GridMap.GetBestClusterNode() != null)
                {
                    Logger.Log(TrinityLogLevel.Debug, LogCategory.Targetting,
                        "| =>   ClusterNode    : {0}",
                        GridMap.GetBestClusterNode().ClusterWeightInfos + " Dist:" + GridMap.GetBestClusterNode().Distance);
                }

                if (MainGrid.Map.Any() && GridMap.GetBestMoveNode() != null)
                {
                    Logger.Log(TrinityLogLevel.Debug, LogCategory.Targetting,
                        "| =>   MoveNode       : {0}",
                        GridMap.GetBestMoveNode().WeightInfos + " Dist:" + GridMap.GetBestMoveNode().Distance);
                }
            }

            return status;
        }

        private static void CheckAvoidDeath()
        {
            using (new MemorySpy("HandleTarget().CheckAvoidDeath"))
            {
                if (Settings.Combat.Misc.AvoidDeath &&
                    Player.CurrentHealthPct <= CombatBase.EmergencyHealthPotionLimit &&
                    !SNOPowerUseTimer(SNOPower.DrinkHealthPotion) &&
                    TargetUtil.AnyMobsInRange(90f, false))
                {
                    Logger.LogNormal("Attempting to avoid death!");
                    Trinity.Player.AvoidDeath = true;
                }
                else
                {
                    Trinity.Player.AvoidDeath = false;
                }
            }
        }

        private static bool HandleTownRunWaitTask()
        {
            while (CurrentTarget == null && (ForceVendorRunASAP || WantToTownRun) && !Zeta.Bot.Logic.BrainBehavior.IsVendoring && TownRun.TownRunTimerRunning())
            {
                Logger.Log(TrinityLogLevel.Info, LogCategory.Behavior, "CurrentTarget is null but we are ready to to Town Run, waiting... ");
                return true;
            }

            while (CurrentTarget == null && Trinity.IsTryingToTownPortal && TownRun.TownRunTimerRunning())
            {
                Logger.Log(TrinityLogLevel.Info, LogCategory.Behavior, "Waiting for town run... ");
                return true;
            }

            return false;
        }

        private static bool HandleTownRunReadyTask()
        {
            if (CurrentTarget == null && Trinity.IsTryingToTownPortal && TownRun.TownRunTimerFinished())
            {
                Logger.Log(TrinityLogLevel.Info, LogCategory.Behavior, "Town Run Ready!");
                return true;
            }
            
            return false;
        }

        private static bool HandleCastingOrLoadingTask()
        {
            return (CurrentTarget == null || !CurrentTarget.IsItem) &&
                Trinity.IsTryingToTownPortal && Player.CurrentHealthPct >= 35 && Player.IsCastingOrLoading;
        }

        private static bool HandleObjectInRange()
        {
            if (CurrentTarget == null)
                return false;

            if (CurrentTarget.IsAvoidance)
            {
                TrinityPower power = AbilitySelector();
                if (!CombatBase.IsNull(power))
                    CombatBase.Cast(power);
            }

            bool stuckOnTarget = false;
            switch (CurrentTarget.Type)
            {
                case GObjectType.Door:
                case GObjectType.HealthWell:
                case GObjectType.Shrine:
                case GObjectType.Container:
                case GObjectType.Interactable:
                case GObjectType.CursedChest:
                case GObjectType.CursedShrine:
                case GObjectType.Destructible:
                case GObjectType.Barricade:
                    if (!ZetaDia.Me.Movement.IsMoving && CurrentTarget.RadiusDistance <= 5f)
                        stuckOnTarget = true;
                    break;
            }

            bool npcInRange = CurrentTarget.IsQuestGiver && CurrentTarget.RadiusDistance <= 3f;

            bool noRangeRequired = CurrentTarget.RequiredRange <= 1f;
            switch (CurrentTarget.Type)
            {
                case GObjectType.Door:
                case GObjectType.Barricade:
                case GObjectType.Destructible:
                    noRangeRequired = false;
                    break;
            }

            if (noRangeRequired || TargetIsInRange || stuckOnTarget || npcInRange)
            {
                Logger.Log(LogCategory.Behavior, "Object in Range: noRangeRequired={0} Target In Range={1} stuckOnTarget={2} npcInRange={3}",
                    noRangeRequired, TargetIsInRange, stuckOnTarget, npcInRange);

                UpdateStatusTextTarget(true);

                switch (CurrentTarget.Type)
                {
                    case GObjectType.Avoidance:
                        break;
                    case GObjectType.Player:
                        break;
                    // Unit, use our primary power to attack
                    case GObjectType.Unit:
                        {
                            if (!CombatBase.IsNull(CombatBase.CurrentPower))
                            {
                                return HandleUnitInRange();
                            }
                            return false;
                        }
                    // Item, interact with it and log item stats
                    case GObjectType.Item:
                        {
                            // Check if we actually have room for this item first
                            Logger.LogDebug(LogCategory.Behavior, "Entering HandleItem");
                            bool isTwoSlot = true;
                            if (CurrentTarget.Item != null && CurrentTarget.Item.CommonData != null)
                            {
                                isTwoSlot = CurrentTarget.Item.CommonData.IsTwoSquareItem;
                            }

                            Vector2 validLocation = TrinityItemManager.FindValidBackpackLocation(isTwoSlot);
                            if (validLocation.X < 0 || validLocation.Y < 0)
                            {
                                Logger.Log("No more space to pickup item, town-run requested at next free moment. (HandleTarget)");
                                ForceVendorRunASAP = true;

                                // Record the first position when we run out of bag space, so we can return later
                                TownRun.SetPreTownRunPosition();
                            }
                            else
                            {
                                return HandleItemInRange() > 0;
                            }

                            return false;
                        }
                    // * Gold & Globe - need to get within pickup radius only
                    case GObjectType.Gold:
                    case GObjectType.HealthGlobe:
                    case GObjectType.PowerGlobe:
                    case GObjectType.ProgressionGlobe:
                        {
                            Logger.LogDebug(LogCategory.Behavior, "Entering Handle Pickup");
                            int interactAttempts;
                            // Count how many times we've tried interacting
                            if (!CacheData.InteractAttempts.TryGetValue(CurrentTarget.RActorGuid, out interactAttempts))
                            {
                                CacheData.InteractAttempts.Add(CurrentTarget.RActorGuid, 1);
                            }
                            else
                            {
                                CacheData.InteractAttempts[CurrentTarget.RActorGuid]++;
                            }
                            // If we've tried interacting too many times, blacklist this for a while
                            if (interactAttempts > 15)
                            {
                                Trinity.TimedBlacklist.Add(CurrentTarget.RActorGuid, Trinity.IgnoreDuringSeconds(15));
                            }

                            return true;
                        }
                    case GObjectType.Door:
                    case GObjectType.HealthWell:
                    case GObjectType.Shrine:
                    case GObjectType.Container:
                    case GObjectType.Interactable:
                    case GObjectType.CursedChest:
                    case GObjectType.CursedShrine:
                        {
                            Logger.LogDebug(LogCategory.Behavior, "Entering Handle Interact");

                            if (ZetaDia.Me.Movement.SpeedXY > 0.5)
                            {
                                Logger.LogVerbose(LogCategory.Behavior, "Trying to stop, Speeds:{0:0.00}/{1:0.00}", ZetaDia.Me.Movement.SpeedXY, PlayerMover.GetMovementSpeed());
                                Navigator.PlayerMover.MoveStop();

                            }
                            else
                            {
                                if (SpellHistory.TimeSinceUse(SNOPower.Axe_Operate_Gizmo) < TimeSpan.FromMilliseconds(150))
                                {
                                    break;
                                }

                                int attemptCount;
                                CacheData.InteractAttempts.TryGetValue(CurrentTarget.RActorGuid, out attemptCount);

                                Logger.LogDebug(LogCategory.UserInformation, "Interacting with {1} Distance {2:0} Radius {3:0.0} Attempt {4}",
                                            SNOPower.Axe_Operate_Gizmo, CurrentTarget.InternalName, CurrentTarget.Distance, CurrentTarget.Radius, attemptCount);

                                if (CurrentTarget.ActorType == ActorType.Monster)
                                    ZetaDia.Me.UsePower(SNOPower.Axe_Operate_NPC, Vector3.Zero, Player.WorldDynamicID, CurrentTarget.ACDGuid);
                                else
                                    ZetaDia.Me.UsePower(SNOPower.Axe_Operate_Gizmo, Vector3.Zero, 0, CurrentTarget.ACDGuid);

                                SpellHistory.RecordSpell(new TrinityPower()
                                {
                                    SNOPower = SNOPower.Axe_Operate_Gizmo,
                                    TargetACDGUID = CurrentTarget.ACDGuid,
                                    MinimumRange = CurrentTarget.RequiredRange,
                                    TargetPosition = CurrentTarget.Position,
                                });

                                // Count how many times we've tried interacting
                                if (!CacheData.InteractAttempts.TryGetValue(CurrentTarget.RActorGuid, out attemptCount))
                                {
                                    CacheData.InteractAttempts.Add(CurrentTarget.RActorGuid, 1);
                                }
                                else
                                {
                                    CacheData.InteractAttempts[CurrentTarget.RActorGuid]++;
                                }

                                return true;
                            }
                            return false;
                        }
                    // * Destructible - need to pick an ability and attack it
                    case GObjectType.Destructible:
                    case GObjectType.Barricade:
                        {
                            Logger.LogDebug(LogCategory.Behavior, "Entering Handle Destroy");
                            if (CombatBase.CurrentPower.SNOPower != SNOPower.None)
                            {
                                // obsolet ?
                                if (CurrentTarget.Type == GObjectType.Barricade)
                                {
                                    Logger.Log(TrinityLogLevel.Debug, LogCategory.Targetting,
                                        "Barricade: Name={0}. SNO={1}, Range={2}. Needed range={3}. Radius={4}. Type={5}. Using power={6}",
                                        CurrentTarget.InternalName,     // 0
                                        CurrentTarget.ActorSNO,         // 1
                                        CurrentTarget.Distance,         // 2
                                        CurrentTarget.RequiredRange,            // 3
                                        CurrentTarget.Radius,           // 4
                                        CurrentTarget.Type,             // 5
                                        CombatBase.CurrentPower.SNOPower// 6 
                                        );
                                }
                                else
                                {
                                    Logger.Log(TrinityLogLevel.Debug, LogCategory.Targetting,
                                        "Destructible: Name={0}. SNO={1}, Range={2}. Needed range={3}. Radius={4}. Type={5}. Using power={6}",
                                        CurrentTarget.InternalName,       // 0
                                        CurrentTarget.ActorSNO,           // 1
                                        TargetDistance,            // 2
                                        CurrentTarget.RequiredRange,              // 3 
                                        CurrentTarget.Radius,             // 4
                                        CurrentTarget.Type,               // 5
                                        CombatBase.CurrentPower.SNOPower  // 6
                                        );
                                }

                                if (CurrentTarget.RActorGuid == IgnoredRActorGuid || DataDictionary.DestroyAtLocationIds.Contains(CurrentTarget.ActorSNO))
                                {
                                    // Location attack - attack the Vector3/map-area (equivalent of holding shift and left-clicking the object in-game to "force-attack")
                                    Vector3 vAttackPoint;
                                    if (CurrentTarget.Distance >= 6f)
                                        vAttackPoint = MathEx.CalculatePointFrom(CurrentTarget.Position, Player.Position, 6f);
                                    else
                                        vAttackPoint = CurrentTarget.Position;

                                    vAttackPoint.Z += 1.5f;
                                    CombatBase.CurrentPower.TargetPosition = vAttackPoint;

                                    if (CombatBase.Cast(CombatBase.CurrentPower))
                                    {
                                        Logger.Log(TrinityLogLevel.Debug, LogCategory.Behavior, "(NB: Attacking location of destructable)");
                                    }
                                }
                                else if (CombatBase.Cast(CombatBase.CurrentPower))
                                {
                                    Logger.Log(TrinityLogLevel.Debug, LogCategory.Behavior, "(NB: Standard attack on ACDGuid)");
                                }

                                int interactAttempts;
                                // Count how many times we've tried interacting
                                if (!CacheData.InteractAttempts.TryGetValue(CurrentTarget.RActorGuid, out interactAttempts))
                                {
                                    CacheData.InteractAttempts.Add(CurrentTarget.RActorGuid, 1);
                                }
                                else
                                {
                                    CacheData.InteractAttempts[CurrentTarget.RActorGuid]++;
                                }

                                CacheData.AbilityLastUsed[CombatBase.CurrentPower.SNOPower] = DateTime.UtcNow;

                                return true;
                            }
                            // Now tell Trinity to get a new target!

                            return false;
                        }
                    default:
                        {
                            Logger.LogError("Default handle target in range encountered for {0} Type: {1}", CurrentTarget.InternalName, CurrentTarget.Type);

                            return false;
                        }
                }
            }

            return false;
        }

        /// <summary>
        /// Checks to see if we need a new monster power and will assign it to <see cref="CurrentPower"/>, distinguishes destructables/barricades from units
        /// </summary>
        private static void HandleAssignAbilityTask()
        {
            using (new PerformanceLogger("HandleTarget.AssignMonsterTargetPower"))
            {
                if (CurrentTarget == null)
                    return;

                // Just in case
                if (CombatBase.IsNull(CombatBase.CurrentPower))
                    ShouldPickNewAbilities = true;

                // force to execute special combat routine like wd soulharverster
                if (DateTime.UtcNow.Subtract(LastPickNewAbilitiesForced).TotalMilliseconds >= 350)
                {
                    ShouldPickNewAbilities = true;
                    LastPickNewAbilitiesForced = DateTime.UtcNow;
                }

                // Find a valid ability if the target is a monster
                if (ShouldPickNewAbilities)
                {
                    ShouldPickNewAbilities = false;
                    LastPickNewAbilitiesForced = DateTime.UtcNow;

                    CombatBase.CurrentPower = AbilitySelector();
                    return;
                }
            }
        }
        /// <summary>
        /// Use a Potion if needed
        /// </summary>
        private static bool UsePotionIfNeededTask()
        {
            using (new PerformanceLogger("HandleTarget.UseHealthPotionIfNeeded"))
            {
                if (!Player.IsIncapacitated && Player.CurrentHealthPct > 0 && SpellHistory.TimeSinceUse(SNOPower.DrinkHealthPotion) > TimeSpan.FromSeconds(30) &&
                    Player.CurrentHealthPct <= CombatBase.EmergencyHealthPotionLimit)
                {
                    var legendaryPotions = CacheData.Inventory.Backpack.Where(i => i.InternalName.ToLower()
                        .Contains("healthpotion_legendary_")).ToList();


                    int dynamicId;
                    if (legendaryPotions.Any())
                    {
                        Logger.Log(TrinityLogLevel.Info, LogCategory.UserInformation, "Using Legendary Potion", 0);
                        dynamicId = legendaryPotions.FirstOrDefault().DynamicId;
                        ZetaDia.Me.Inventory.UseItem(dynamicId);
                        SpellHistory.RecordSpell(new TrinityPower(SNOPower.DrinkHealthPotion));
                        return true;
                    }

                    ACDItem potion = ZetaDia.Me.Inventory.BaseHealthPotion;
                    if (potion != null)
                    {
                        int id = potion.DynamicId;
                        Logger.Log(TrinityLogLevel.Info, LogCategory.UserInformation, "Using Regular Potion", 0);
                        ZetaDia.Me.Inventory.UseItem(id);

                        SpellHistory.RecordSpell(new TrinityPower(SNOPower.DrinkHealthPotion));
                        return true;
                    }

                    Logger.Log(TrinityLogLevel.Info, LogCategory.UserInformation, "No Available potions!", 0);
                }
                return false;
            }
        }

        /// <summary>
        /// If we can use special class movement abilities, this will use it and return true
        /// </summary>
        /// <returns></returns>
        private static bool UsedSpecialMovement()
        {
            bool Monk_SpecialMovement = ((CurrentTarget.Type == GObjectType.Gold ||
                CurrentTarget.IsUnit ||
                CurrentTarget.Type == GObjectType.Barricade ||
                CurrentTarget.Type == GObjectType.Destructible) && MonkCombat.IsTempestRushReady());

            // If we're doing avoidance, globes or backtracking, try to use special abilities to move quicker
            if ((CurrentTarget.Type == GObjectType.Avoidance ||
                CurrentTarget.Type == GObjectType.HealthGlobe ||
                CurrentTarget.Type == GObjectType.PowerGlobe ||
                CurrentTarget.Type == GObjectType.ProgressionGlobe ||
                Monk_SpecialMovement)
                && NavHelper.CanRayCast(MovePosition)
                )
            {
                bool attackableSpecialMovement = CurrentTarget.Type == GObjectType.Avoidance &&
                ObjectCache.Any(u => (u.IsUnit || u.Type == GObjectType.Destructible || u.Type == GObjectType.Barricade) &&
                    MathUtil.IntersectsPath(u.Position, u.Radius, Player.Position, CurrentTarget.Position));

                // Leap movement for a barb
                if (CombatBase.CanCast(SNOPower.Barbarian_Leap))
                {
                    return CombatBase.Cast(new TrinityPower(SNOPower.Barbarian_Leap, 0f, MovePosition));
                }

                // Furious Charge movement for a barb
                if (CombatBase.CanCast(SNOPower.Barbarian_FuriousCharge) && attackableSpecialMovement)
                {
                    return CombatBase.Cast(new TrinityPower(SNOPower.Barbarian_FuriousCharge, 0f, MovePosition));
                }

                // Teleport for a wizard (need to be able to check skill rune in DB for a 3-4 teleport spam in a row)
                if (CombatBase.CanCast(SNOPower.Wizard_Teleport))
                {
                    return CombatBase.Cast(new TrinityPower(SNOPower.Wizard_Teleport, 0f, MovePosition));
                }

                // Archon Teleport for a wizard (need to be able to check skill rune in DB for a 3-4 teleport spam in a row)
                if (CombatBase.CanCast(SNOPower.Wizard_Archon_Teleport))
                {
                    return CombatBase.Cast(new TrinityPower(SNOPower.Wizard_Archon_Teleport, 0f, MovePosition));
                }

                // Tempest rush for a monk
                if (CombatBase.CanCast(SNOPower.Monk_TempestRush) && Player.PrimaryResource >= Settings.Combat.Monk.TR_MinSpirit &&
                    ((CurrentTarget.Type == GObjectType.Item && CurrentTarget.Distance > 20f) || CurrentTarget.Type != GObjectType.Item) &&
                    Settings.Combat.Monk.TROption != TempestRushOption.MovementOnly &&
                    MonkCombat.IsTempestRushReady())
                {
                    if (CombatBase.Cast(new TrinityPower(SNOPower.Monk_TempestRush, 0f, MovePosition)))
                    {
                        LastMoveToTarget = MovePosition;
                        return true;
                    }
                }

                // Strafe for a Demon Hunter
                if (attackableSpecialMovement && CombatBase.CanCast(SNOPower.DemonHunter_Strafe))
                {
                    if (CombatBase.Cast(new TrinityPower(SNOPower.DemonHunter_Strafe, 0f, MovePosition)))
                    {
                        LastMoveToTarget = MovePosition;
                        return true;
                    }
                }
            }

            return false;
        }

        /// <summary>
        /// Updates bot status text with appropriate information if we are moving into range of our <see cref="CurrentTarget"/>
        /// </summary>
        private static string lastStatusText = "";
        private static void UpdateStatusTextTarget(bool targetIsInRange)
        {
            string action = "";

            StringBuilder statusText = new StringBuilder();
            if (!targetIsInRange)
                action = "Moveto ";
            else
                switch (CurrentTarget.Type)
                {
                    case GObjectType.Avoidance:
                        action = "Avoid ";
                        break;
                    case GObjectType.Unit:
                        action = "Attack ";
                        break;
                    case GObjectType.Item:
                    case GObjectType.Gold:
                    case GObjectType.PowerGlobe:
                    case GObjectType.HealthGlobe:
                    case GObjectType.ProgressionGlobe:
                        action = "Pickup ";
                        break;
                    case GObjectType.Interactable:
                        action = "Interact ";
                        break;
                    case GObjectType.Door:
                    case GObjectType.Container:
                        action = "Open ";
                        break;
                    case GObjectType.Destructible:
                    case GObjectType.Barricade:
                        action = "Destroy ";
                        break;
                    case GObjectType.Shrine:
                        action = "Click ";
                        break;
                }
            statusText.Append(action);

            statusText.Append("Target=");
            statusText.Append(CurrentTarget.InternalName);
            if (CurrentTarget.IsUnit && CombatBase.CurrentPower.SNOPower != SNOPower.None)
            {
                statusText.Append(" Power=");
                statusText.Append(CombatBase.CurrentPower.SNOPower);
            }
            statusText.Append(" Speed=");
            statusText.Append(ZetaDia.Me.Movement.SpeedXY.ToString("0.00"));
            statusText.Append(" SNO=");
            statusText.Append(CurrentTarget.ActorSNO.ToString(CultureInfo.InvariantCulture));
            statusText.Append(" Elite=");
            statusText.Append(CurrentTarget.IsBossOrEliteRareUnique.ToString());
            statusText.Append(" Weight=");
            statusText.Append(CurrentTarget.Weight.ToString("0"));
            statusText.Append(" Type=");
            statusText.Append(CurrentTarget.Type.ToString());
            statusText.Append(" C-Dist=");
            statusText.Append(CurrentTarget.Distance.ToString("0.0"));
            statusText.Append(" R-Dist=");
            statusText.Append(CurrentTarget.RadiusDistance.ToString("0.0"));
            statusText.Append(" RangeReq'd=");
            statusText.Append(CurrentTarget.RequiredRange.ToString("0.0"));
            statusText.Append(" DistfromTrgt=");
            statusText.Append(TargetDistance.ToString("0"));
            statusText.Append(" tHP=");
            statusText.Append((CurrentTarget.HitPointsPct * 100).ToString("0"));
            statusText.Append(" MyHP=");
            statusText.Append((Player.CurrentHealthPct * 100).ToString("0"));
            statusText.Append(" MyMana=");
            statusText.Append((Player.PrimaryResource).ToString("0"));
            statusText.Append(" InLoS=");
            statusText.Append(TargetIsInLoS.ToString());

            if (Settings.Advanced.DebugInStatusBar)
            {
                TrinityStatusText = statusText.ToString();
                BotMain.StatusText = TrinityStatusText;
            }
            if (lastStatusText != statusText.ToString())
            {
                // prevent spam
                lastStatusText = statusText.ToString();
                Logger.Log(TrinityLogLevel.Debug, LogCategory.Targetting, "{0}", statusText.ToString());
                ResetTrinityStatusText = true;
            }
        }

        /// <summary>
        /// Moves our player if no special ability is available
        /// </summary>
        private static void SetQueuedSpecialMovement()
        {
            if (CurrentTarget == null)
                return;

            if (CurrentTarget.Type == GObjectType.Player ||
                CurrentTarget.Type == GObjectType.OocAvoidance)
            {
                Navigator.PlayerMover.MoveStop();
                return;
            }

            try
            {
                // Moving to kite point
                if (CurrentTarget.IsKite)
                {
                    PlayerMover.SetKiteMovement(CurrentTarget.InternalName);
                    CombatBase.QueuedMovement.CurrentMovement.StopCondition += m =>
                        CurrentTarget == null || !CurrentTarget.IsKite;
                }

                // Moving to Avoidance point
                if (CurrentTarget.IsAvoidance)
                {
                    PlayerMover.SetKiteMovement(CurrentTarget.InternalName);
                    CombatBase.QueuedMovement.CurrentMovement.StopCondition += m =>
                        CurrentTarget == null || !CurrentTarget.IsAvoidance;
                }

                // Keep kiting option
                if (Settings.Combat.Misc.KeepMovingInCombat && Player.IsRanged && CurrentTarget.IsUnit &&
                    TargetIsInRange && !CombatBase.IsNull(CombatBase.CurrentPower) &&
                    MainGrid.Map.Any() &&
                    (Player.IsRanged || TargetUtil.ClusterExists(Settings.Combat.Misc.TrashPackClusterRadius, TargetDistance + 5f, Settings.Combat.Misc.TrashPackSize)))
                {
                    CombatBase.QueuedMovement.Queue(new QueuedMovement
                    {
                        Id = CurrentTarget.RActorGuid,
                        Name = CurrentTarget.InternalName,
                        Infos = "(Keep moving) " + CurrentTarget.Infos,
                        Destination = GridMap.GetBestMoveNode().Position,
                        OnUpdate = m =>
                        {
                            var _safeNode = GridMap.GetBestMoveNode();
                            if (_safeNode != null)
                                m.Destination = _safeNode.Position;
                        },
                        StopCondition = m =>
                            m.Destination == Vector3.Zero ||
                            m.Destination.Distance2D(Trinity.Player.Position) <= 1f ||
                            CurrentTarget == null ||
                            CombatBase.IsNull(CombatBase.CurrentPower) || CurrentTarget.IsAvoidance ||
                            !CurrentTarget.IsUnit || (CurrentTarget.IsUnit && !TargetIsInRange)
                        ,
                        Options = new QueuedMovementOptions
                        {
                            Logging = LogLevel.Info,
                            AcceptableDistance = 1f,
                            Type = MoveType.KeepMoving,
                        }
                    });
                }

                if (CurrentTarget.IsUnit && !CombatBase.IsNull(CombatBase.CurrentPower) && CombatBase.CurrentPower.SNOPower == SNOPower.Walk && CombatBase.CurrentPower.TargetPosition != Vector3.Zero)
                {
                    CombatBase.QueuedMovement.Queue(new QueuedMovement
                    {
                        Id = CurrentTarget.RActorGuid,
                        Name = CurrentTarget.InternalName,
                        Infos = "(Walk) " + CurrentTarget.Infos,
                        Destination = CombatBase.CurrentPower.TargetPosition,
                        StopCondition = m =>
                            m.Destination == Vector3.Zero ||
                            m.Destination.Distance2D(Trinity.Player.Position) <= 1f ||
                            CurrentTarget == null ||
                            CombatBase.IsNull(CombatBase.CurrentPower) || CurrentTarget == null || CurrentTarget.IsAvoidance ||
                            !CurrentTarget.IsUnit || CombatBase.CurrentPower.SNOPower != SNOPower.Walk
                        ,
                        Options = new QueuedMovementOptions
                        {
                            Logging = LogLevel.Info,
                            AcceptableDistance = 3f,
                            Type = MoveType.BasicCombat,
                        }
                    });
                }

                if (CombatBase.QueuedMovement.IsQueuedMovement)
                {
                    LastSentMovePower = DateTime.UtcNow;
                    LastMoveToTarget = CombatBase.QueuedMovement.CurrentMovement.Destination;
                }
            }
            catch (Exception ex)
            {
                Logger.LogDebug("Error in HandleTarget.SetQueuedSpecialMovement: " + ex);
            }
        }

        /// <summary>
        /// Moves our player if no special ability is available
        /// </summary>
        /// <param name="bForceNewMovement"></param>
        private static void SetQueuedBasicMovement()
        {
            using (new PerformanceLogger("HandleTarget.HandleBasicMovement"))
            {
                if (CurrentTarget.Type == GObjectType.Player ||
                    CurrentTarget.Type == GObjectType.OocAvoidance)
                {
                    Navigator.PlayerMover.MoveStop();
                    return;
                }
                if (DateTime.UtcNow.Subtract(LastSentMovePower).TotalMilliseconds >= 250 || Vector3.Distance(LastMoveToTarget, MovePosition) >= 2f)
                {
                    if (!CurrentTarget.IsAvoidance && !(CurrentTarget.IsUnit && TargetIsInRange))
                    {
                        CombatBase.QueuedMovement.Queue(new QueuedMovement
                        {
                            Id = CurrentTarget.RActorGuid,
                            Name = CurrentTarget.InternalName,
                            Infos = MovePosition != CurrentTarget.Position ? "(Required destination) " + CurrentTarget.Infos : CurrentTarget.Infos,
                            Destination = MovePosition,
                            OnUpdate = m =>
                            {
                                if (CurrentTarget != null && CurrentTarget.Type == GObjectType.Player || CurrentTarget.Type == GObjectType.OocAvoidance)
                                {
                                    Navigator.PlayerMover.MoveStop();
                                }
                                else if (CurrentTarget != null && !CurrentTarget.IsAvoidance && !TargetIsInRange)
                                {
                                    SetTargetFields();
                                    m.Name = CurrentTarget.InternalName;
                                    m.Infos = MovePosition != CurrentTarget.Position ? "(Required destination) " + CurrentTarget.Infos : CurrentTarget.Infos;
                                    m.Destination = MovePosition;
                                    m.Options.Type = CurrentTarget.IsUnit ? MoveType.BasicCombat : MoveType.TargetAttempt;
                                }
                            },
                            StopCondition = m =>
                                CurrentTarget == null || CurrentTarget.IsAvoidance || CurrentTarget.Type == GObjectType.Player || CurrentTarget.Type == GObjectType.OocAvoidance || TargetIsInRange
                            ,
                            Options = new QueuedMovementOptions
                            {
                                Logging = LogLevel.Info,
                                AcceptableDistance = 2f,
                                Type = CurrentTarget.IsUnit ? MoveType.BasicCombat : MoveType.TargetAttempt
                            }
                        });
                    }

                    if (CombatBase.QueuedMovement.IsQueuedMovement)
                    {
                        LastSentMovePower = DateTime.UtcNow;
                        LastMoveToTarget = MovePosition;
                    }
                }
            }
        }

        private static void SetTargetFields()
        {
            using (new PerformanceLogger("HandleTarget.SetRequiredRange"))
            {
                if (CurrentTarget == null)
                    return;

                CurrentTarget.RequiredRange = 5f;

                TargetDistance = CurrentTarget.RadiusDistance;
                TargetIsInLoS = CurrentTarget.IsInLineOfSight;

                TargetPosition = CurrentTarget.Position;
                MovePosition = CurrentTarget.Position;

                #region switch (CurrentTarget.Type)
                switch (CurrentTarget.Type)
                {
                    // * Unit, we need to pick an ability to use and get within range
                    case GObjectType.Unit:
                        {
                            CurrentTarget.RequiredRange = CombatBase.CurrentPower.MinimumRange;

                            if (CombatBase.CurrentPower.TargetACDGUID != -1)
                            {
                                var target = ObjectCache.FirstOrDefault(u => u.ACDGuid == CombatBase.CurrentPower.TargetACDGUID);
                                if (target != null)
                                {
                                    // This move position is for an aoe movement
                                    // the required range is not checked with this position but with the Acd position 
                                    if (CombatBase.CurrentPower.MovePosition != Vector3.Zero)
                                    {
                                        TargetPosition = CombatBase.CurrentPower.MovePosition;
                                    }
                                    else
                                    {
                                        TargetPosition = target.Position;
                                    }

                                    // Distance to ACD
                                    TargetDistance = Math.Max(target.Position.Distance2D(Player.Position) - target.Radius, 0f);
                                    CombatBase.CurrentPower.TargetPosition = target.Position;
                                }
                            }
                            else if (CombatBase.CurrentPower.TargetPosition != Vector3.Zero)
                            {
                                // This move position is for a required position before cast at the target position
                                if (CombatBase.CurrentPower.MovePosition != Vector3.Zero)
                                {
                                    TargetPosition = CombatBase.CurrentPower.MovePosition;
                                }
                                else
                                {
                                    TargetPosition = CombatBase.CurrentPower.TargetPosition;
                                }

                                // Distance to destination
                                TargetDistance = Math.Max(TargetPosition.Distance2D(Player.Position) - CurrentTarget.Radius, 0f);
                            }

                            MovePosition = TargetPosition;
                            break;
                        }
                    // * Item - need to get within 6 feet and then interact with it
                    case GObjectType.Item:
                        {
                            CurrentTarget.RequiredRange = 6f;
                            TargetDistance = CurrentTarget.Distance;
                            break;
                        }
                    // * Gold - need to get within pickup radius only
                    case GObjectType.Gold:
                        {
                            CurrentTarget.RequiredRange = Math.Max(Player.GoldPickupRadius - 5f, 5f);
                            TargetDistance = CurrentTarget.Distance;
                            MovePosition = MathEx.CalculatePointFrom(Player.Position, CurrentTarget.Position, -2f);
                            break;
                        }
                    // * Globes - need to get within pickup radius only
                    case GObjectType.PowerGlobe:
                    case GObjectType.HealthGlobe:
                    case GObjectType.ProgressionGlobe:
                        {
                            CurrentTarget.RequiredRange = Math.Max(Player.GoldPickupRadius - 5f, 5f);
                            TargetDistance = CurrentTarget.Distance;
                            break;
                        }
                    // * Shrine & Container - need to get within 8 feet and interact
                    case GObjectType.HealthWell:
                        {
                            CurrentTarget.RequiredRange = 4f;

                            float range;
                            if (DataDictionary.CustomObjectRadius.TryGetValue(CurrentTarget.ActorSNO, out range))
                            {
                                CurrentTarget.RequiredRange = range;
                            }
                            break;
                        }
                    case GObjectType.Shrine:
                    case GObjectType.Container:
                        {
                            CurrentTarget.RequiredRange = 6f;

                            float range;
                            if (DataDictionary.CustomObjectRadius.TryGetValue(CurrentTarget.ActorSNO, out range))
                            {
                                CurrentTarget.RequiredRange = range;
                            }
                            if (CurrentTarget.RequiredRange <= 1)
                                CurrentTarget.RequiredRange = 2f;
                            break;
                        }
                    case GObjectType.Interactable:
                        {
                            if (CurrentTarget.IsQuestGiver)
                            {
                                MovePosition = MathEx.CalculatePointFrom(CurrentTarget.Position, Player.Position, CurrentTarget.Radius + 2f);
                                CurrentTarget.RequiredRange = 5f;
                            }
                            else
                            {
                                CurrentTarget.RequiredRange = 5f;
                            }
                            // Check if it's in our interactable range dictionary or not
                            float range;

                            if (DataDictionary.CustomObjectRadius.TryGetValue(CurrentTarget.ActorSNO, out range))
                            {
                                CurrentTarget.RequiredRange = range;
                            }
                            if (CurrentTarget.RequiredRange <= 0)
                                CurrentTarget.RequiredRange = CurrentTarget.Radius;

                            break;
                        }
                    // * Destructible - need to pick an ability and attack it
                    case GObjectType.Destructible:
                        {
                            CurrentTarget.RequiredRange = Math.Max(5f, CombatBase.CurrentPower.MinimumRange);
                            CurrentTarget.Radius = 1f;
                            TargetDistance = CurrentTarget.Distance;
                            break;
                        }
                    case GObjectType.Barricade:
                        {
                            // Pick a range to try to reach + (tmp_fThisRadius * 0.70);
                            CurrentTarget.RequiredRange = Math.Max(5f, CombatBase.CurrentPower.MinimumRange);
                            //CurrentTarget.Radius = 1f;
                            TargetDistance = CurrentTarget.Distance;
                            break;
                        }
                    // * Avoidance - need to pick an avoid location and move there
                    case GObjectType.Avoidance:
                        {
                            CurrentTarget.RequiredRange = 2f;
                            break;
                        }
                    case GObjectType.Door:
                        CurrentTarget.RequiredRange = 5f;
                        TargetDistance = CurrentTarget.Distance;
                        break;
                    default:
                        CurrentTarget.RequiredRange = CurrentTarget.Radius;
                        break;
                }
                #endregion

                TargetIsInRange = CurrentTarget.RequiredRange <= 1f || (TargetDistance <= CurrentTarget.RequiredRange && TargetIsInLoS);
            }
        }

        private static bool HandleUnitInRange()
        {
            Logger.LogDebug(LogCategory.Behavior, "Entering HandleUnit");
            using (new PerformanceLogger("HandleTarget.HandleUnitInRange"))
            {
                bool usePowerResult = CombatBase.Cast(CombatBase.CurrentPower);
                if (usePowerResult)
                    ShouldPickNewAbilities = true;

                return usePowerResult;
            }
        }

        private static int HandleItemInRange()
        {
            using (new PerformanceLogger("HandleTarget.HandleItemInRange"))
            {
                int iInteractAttempts;
                // Pick the item up the usepower way, and "blacklist" for a couple of loops
                ZetaDia.Me.UsePower(SNOPower.Axe_Operate_Gizmo, Vector3.Zero, 0, CurrentTarget.ACDGuid);
                IgnoredRActorGuid = CurrentTarget.RActorGuid;
                IgnoredRActorGuidForLoops = 3;
                // Store item pickup stats

                string itemSha1Hash = HashGenerator.GenerateItemHash(CurrentTarget.Position, CurrentTarget.ActorSNO, CurrentTarget.InternalName, Player.WorldDynamicID, CurrentTarget.ItemQuality, CurrentTarget.ItemLevel);
                if (!ItemDropStats._hashsetItemPicksLookedAt.Contains(itemSha1Hash))
                {
                    ItemDropStats._hashsetItemPicksLookedAt.Add(itemSha1Hash);
                    GItemType itemType = TrinityItemManager.DetermineItemType(CurrentTarget.InternalName, CurrentTarget.DBItemType, CurrentTarget.FollowerType);
                    GItemBaseType itemBaseType = TrinityItemManager.DetermineBaseType(itemType);
                    if (itemBaseType == GItemBaseType.Armor || itemBaseType == GItemBaseType.WeaponOneHand || itemBaseType == GItemBaseType.WeaponTwoHand ||
                        itemBaseType == GItemBaseType.WeaponRange || itemBaseType == GItemBaseType.Jewelry || itemBaseType == GItemBaseType.FollowerItem ||
                        itemBaseType == GItemBaseType.Offhand)
                    {
                        int iQuality;
                        ItemDropStats.ItemsPickedStats.Total++;
                        if (CurrentTarget.ItemQuality >= ItemQuality.Legendary)
                            iQuality = ItemDropStats.QUALITYORANGE;
                        else if (CurrentTarget.ItemQuality >= ItemQuality.Rare4)
                            iQuality = ItemDropStats.QUALITYYELLOW;
                        else if (CurrentTarget.ItemQuality >= ItemQuality.Magic1)
                            iQuality = ItemDropStats.QUALITYBLUE;
                        else
                            iQuality = ItemDropStats.QUALITYWHITE;
                        //asserts	
                        if (iQuality > ItemDropStats.QUALITYORANGE)
                        {
                            Logger.Log(TrinityLogLevel.Error, LogCategory.UserInformation, "ERROR: Item type (" + iQuality + ") out of range");
                        }
                        if ((CurrentTarget.ItemLevel < 0) || (CurrentTarget.ItemLevel >= 74))
                        {
                            Logger.Log(TrinityLogLevel.Error, LogCategory.UserInformation, "ERROR: Item level (" + CurrentTarget.ItemLevel + ") out of range");
                            CurrentTarget.ItemLevel = 0;
                        }
                        ItemDropStats.ItemsPickedStats.TotalPerQuality[iQuality]++;
                        ItemDropStats.ItemsPickedStats.TotalPerLevel[CurrentTarget.ItemLevel]++;
                        ItemDropStats.ItemsPickedStats.TotalPerQPerL[iQuality, CurrentTarget.ItemLevel]++;
                    }
                    else if (itemBaseType == GItemBaseType.Gem)
                    {
                        int iGemType = 0;
                        ItemDropStats.ItemsPickedStats.TotalGems++;
                        if (itemType == GItemType.Topaz)
                            iGemType = ItemDropStats.GEMTOPAZ;
                        if (itemType == GItemType.Ruby)
                            iGemType = ItemDropStats.GEMRUBY;
                        if (itemType == GItemType.Emerald)
                            iGemType = ItemDropStats.GEMEMERALD;
                        if (itemType == GItemType.Amethyst)
                            iGemType = ItemDropStats.GEMAMETHYST;
                        if (itemType == GItemType.Diamond)
                            iGemType = ItemDropStats.GEMDIAMOND;

                        ItemDropStats.ItemsPickedStats.GemsPerType[iGemType]++;
                        ItemDropStats.ItemsPickedStats.GemsPerLevel[CurrentTarget.ItemLevel]++;
                        ItemDropStats.ItemsPickedStats.GemsPerTPerL[iGemType, CurrentTarget.ItemLevel]++;
                    }
                    else if (itemType == GItemType.HealthPotion)
                    {
                        ItemDropStats.ItemsPickedStats.TotalPotions++;
                        if ((CurrentTarget.ItemLevel < 0) || (CurrentTarget.ItemLevel > 63))
                        {
                            Logger.Log(TrinityLogLevel.Error, LogCategory.UserInformation, "ERROR: Potion level ({0}) out of range", CurrentTarget.ItemLevel);
                        }
                        ItemDropStats.ItemsPickedStats.PotionsPerLevel[CurrentTarget.ItemLevel]++;
                    }
                    else if (c_item_GItemType == GItemType.InfernalKey)
                    {
                        ItemDropStats.ItemsPickedStats.TotalInfernalKeys++;
                    }
                }

                // Count how many times we've tried interacting
                if (!CacheData.InteractAttempts.TryGetValue(CurrentTarget.RActorGuid, out iInteractAttempts))
                {
                    CacheData.InteractAttempts.Add(CurrentTarget.RActorGuid, 1);

                    // Fire item looted for Demonbuddy Item stats
                    GameEvents.FireItemLooted(CurrentTarget.ACDGuid);
                }
                else
                {
                    CacheData.InteractAttempts[CurrentTarget.RActorGuid]++;
                }
                // If we've tried interacting too many times, blacklist this for a while
                if (iInteractAttempts > 35 && CurrentTarget.ItemQuality < ItemQuality.Legendary)
                {
                    //Blacklist60Seconds.Add(CurrentTarget.RActorGuid);
                }
                // Now tell Trinity to get a new target!
                return iInteractAttempts;
            }
        }
    }
}
